;Game of Life - Cellular Automata Simulation
;16x16 pixels LED display consisting of four 8x8 dot matrix modules

;Rules:
;  less than two neighbours: cell dies
;  two neighbours: cell maintains current state
;  three neighbours: cell comes to life
;  four or more neighbours: cell dies

;ATTiny2313 @ 8 MHz (internal R/C oscillator)
;Fuses:  low = 0xC4 (8 MHz internal osc., BOD enabled)
;       high = 0xD9 (BOD 4.3V)
.include "../include/tn2313def.inc"

.equ	FREQ = 8000000 ;clock frequency
.equ	FPS = 5 ;frames per second

;PORTS:
;  PB0-7 = Keys 1-8
;  PD0   = RS232 Rx
;  PD1   = RS232 Tx
;  PD2   = Cols Data
;  PD3   = Cols Clock
;  PD4   = Rows Data
;  PD5   = Rows Clock
;  PD6   = Strobe (Cols & Rows)
.equ	ColDa = 2
.equ	ColCk = 3
.equ	RowDa = 4
.equ	RowCk = 5
.equ	Strobe = 6

;VERSION HISTORY
;  Version 0.1  2007-02-02  Arne Rossius
;  Version 0.5  2007-02-13  Arne Rossius
;  Version 0.6  2007-07-09  Arne Rossius
;    * added pattern selection by keys

.def	left = R0
.def	right = R1
.def	row = R2

.def	temp = R16
.def	temp2 = R17
.def	byte = R18
.def	speed = R19
.def	x = R20
.def	y = R21
.def	rxbuf = R22

.def	flags = R25
	.equ	NextFrame = 0

;R29:R28 = Y pointer (RAM)
;R31:R30 = Z pointer (Flash, used by RS232 Tx interrupt and the main program)

.equ	SpeedDiv = FREQ/8/256/16/FPS

;===============================================================================

.dseg
	array1: .byte 32
	array2: .byte 32
	buffer: .byte 48
	stack: .byte 16
	
;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	.db SpeedDiv
.org OVF0addr
	rjmp	t0ovf
	
;===============================================================================

.org PAGESIZE*2 ;align to flash page boundary
firstframes:
	;#1: fountain
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x06, 0xC0, 0x04, 0x40, 0x12, 0x90
	.db 0x14, 0x50, 0x08, 0x20, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

	;#2: "Queen Bee Shutle" (doesn't work as expected on a 16x16 grid) -> pulsar
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x01, 0x80, 0x00, 0x40, 0x00, 0x20, 0x00, 0x20
	.db 0x00, 0x20, 0x00, 0x40, 0x01, 0x80, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	
	;#3: loads of action -> block + glider -> 4 oscillators
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x1C, 0x00, 0x08, 0x00, 0x1E, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

	;#4: "R-pentomino" -> oscillator
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x01, 0x80, 0x03, 0x00
	.db 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	
	;#5: infinity sign -> "Pulsar" (4-quadrant 3-cycle oscillator)
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x06, 0xC0, 0x09, 0x20
	.db 0x06, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

	;#6: -> dies
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x10, 0x00
	.db 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

	;#7: diagonal glider
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x40, 0x00, 0x20, 0x00, 0xe0, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	
	;#8: blocks and a blinker (test)
	.db 0x00, 0x03, 0x00, 0x03, 0x00, 0x30, 0x00, 0x30
	.db 0x30, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00
	.db 0x00, 0x00, 0x00, 0x70, 0x30, 0x00, 0x30, 0x00
	.db 0x00, 0x03, 0x00, 0x03, 0x06, 0x00, 0x06, 0x00
	
;-------------------------------------------------------------------------------

t0ovf:	;timer 0 overflow
	push	temp
	push	temp2
	push	YL
	push	YH
	in	temp, SREG
	push	temp
	
	;multiplex display
	inc	row
	ldi	temp, 0x0F
	and	row, temp
	;shift column
	ldi	YL, array1
	add	YL, row
	add	YL, row
	ld	temp, Y+
	rcall	shift
	ld	temp, Y+
	rcall	shift
	;select row
	ldi	temp2, 15
rowsel_loop:
	sbi	PORTD, RowDa
	cp	row, temp2
	brne	PC+2
	cbi	PORTD, RowDa
	sbi	PORTD, RowCk
	cbi	PORTD, RowCk
	subi	temp2, 1
	brcc	rowsel_loop
	;latch new data
	sbi	PORTD, Strobe
	cbi	PORTD, Strobe
	
	;set "NextFrame" flag every speed-th cycle
	tst	row
	brne	ovf0_end
	dec	speed
	brne	ovf0_end
	ldi	speed, SpeedDiv
	ori	flags, 1<<NextFrame
	
ovf0_end:
	pop	temp
	out	SREG, temp
	pop	YH
	pop	YL
	pop	temp2
	pop	temp
	reti

;-------------------------------------------------------------------------------

reset:
	;stack pointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;ports
	ldi	temp, 0xFF
	out	PORTB, temp
	ldi	temp, 0x7F
	out	DDRD, temp
	
	;timer 0
	ldi	temp, 0x02 ;Clk/8
	out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;RS232
	ldi	temp, 1<<TXEN | 1<<RXEN | 1<<RXCIE
	out	UCSRB, temp
	ldi	temp, 12 ;38400 Baud @ 8 MHz
	out	UBRRL, temp
	
	;initialize registers
	clr	row
	clr	rxbuf
	ldi	speed, SpeedDiv
	ldi	flags, 1<<NextFrame
	
	;load first frame from flash to array2
	ldi	ZH, HIGH(firstframes*2)
	ldi	ZL, LOW(firstframes*2)
	ldi	YL, array2
	rcall	flash2ram
	;enable interrupts
	sei

;-------------------------------------------------------------------------------

main:
	;load new pattern when a key is pressed
	in	temp, PINB
	ldi	temp2, 8
	ldi	ZL, LOW(firstframes*2)
	ldi	ZH, HIGH(firstframes*2)
keys_loop:
	sbrc	temp, 0
	rjmp	keys_loop_next
	push	temp2
	ldi	YL, array2
	rcall	flash2ram
	ori	flags, 1<<NextFrame
	pop	temp2
keys_loop_next:
	adiw	ZH:ZL, 32
	lsr	temp
	dec	temp2
	brne	keys_loop
	
	;check for RS232 reception
;	sbic	USR, RXC
;	rjmp	rs232_rx
	
	;wait until the time has come to display the next frame
	sbrs	flags, NextFrame
	rjmp	main

	;copy array2 to array1 (display new frame)
	ldi	temp2, 32
	ldi	YL, array1
copy_loop:
	ldd	temp, Y+(array2-array1)
	st	Y+, temp
	dec	temp2
	brne	copy_loop
	
	;calculate next frame
	ldi	y, 15
y_loop:	
	ldi	x, 15
x_loop:
	;find number of live neighbours
	clr	temp
	dec	x
	andi	x, 0x0F
	rcall	getcell ;(x-1 | y)
	dec	y
	andi	y, 0x0F
	rcall	getcell ;(x-1 | y-1)
	inc	x
	andi	x, 0x0F
	rcall	getcell ;(x | y-1)
	inc	x
	andi	x, 0x0F
	rcall	getcell ;(x+1 | y-1)
	inc	y
	andi	y, 0x0F
	rcall	getcell ;(x+1 | y)
	inc	y
	andi	y, 0x0F
	rcall	getcell ;(x+1 | y+1)
	dec	x
	andi	x, 0x0F
	rcall	getcell ;(x | y+1)
	dec	x
	andi	x, 0x0F
	rcall	getcell ;(x-1 | y+1)
	inc	x
	andi	x, 0x0F
	dec	y
	andi	y, 0x0F
	;find next state
	cpi	temp, 2 ;<2 neighbours: cell dies
	brlo	die
	cpi	temp, 3 ;3 neighbours: cell comes to life
	breq	new
	cpi	temp, 4 ;>= 4 neighbours: cell dies
	brsh	die
	rcall	getcell ;2 neighbours: cell maintains current state
	lsr	byte ;copy current state to CARRY flag
	rjmp	calc_end
die:
	clc
	rjmp	calc_end
new:
	sec
calc_end:
	ror	left ;save calculated bit
	ror	right
	subi	x, 1 ;x loop
	brcc	x_loop
	ldi	YL, array2 ;save calculated line to array2
	add	YL, y
	add	YL, y
	st	Y+, left
	st	Y+, right
	subi	y, 1 ;y loop
	brcc	y_loop
	
	;reset "NextFrame" flag
	andi	flags, ~(1<<NextFrame)
	
	rjmp	main
	
;--------------------

getcell:
	;add the state of a cell in array1 to "temp" and return it in bit 0 of "byte"
	ldi	YL, array1
	mov	temp2, x
	lsr	temp2 ;x = x / 8
	lsr	temp2
	lsr	temp2
	add	YL, temp2 ;offset x / 8
	add	YL, y
	add	YL, y ;offset y * 2
	ld	byte, Y
	mov	temp2, x
	andi	temp2, 0x07
getcell_shift:
	cpi	temp2, 7
	breq	getcell_end
	lsr	byte
	inc	temp2
	rjmp	getcell_shift
getcell_end:
	sbrc	byte, 0
	inc	temp
	ret
	
;--------------------

flash2ram:
	;copy 32 bytes from ZH:ZL (Flash) to YL (RAM)
	ldi	temp2, 32
flash2ram_loop:
	lpm	byte, Z+
	st	Y+, byte
	dec	temp2
	brne	flash2ram_loop
	sbiw	ZH:ZL, 32
	ret
	
;-------------------------------------------------------------------------------

shift:
	;output "temp" to columns shift register (MSB first)
	ldi	temp2, 8
shift_loop:
	cbi	PORTD, ColDa
	lsl	temp
	brcc	PC+2
	sbi	PORTD, ColDa
	sbi	PORTD, ColCk
	cbi	PORTD, ColCk
	dec	temp2
	brne	shift_loop
	ret

;===============================================================================

;DATA area for RS232 transmit strings
mainmenu:
	.db 0x0d, 0x0a
	.db "GAME OF LIFE - MAIN MENU", 0x0d, 0x0a
	.db 0x0d, 0x0a
	.db "1) upload new starting pattern", 0x0d, 0x0a
	.db "2) download a starting pattern", 0x0d, 0x0a
	.db "3) change speed ", 0x0d, 0x0a
	.db 0x0d, 0x0a
	.db "Enter your choice: ", 0x00

number:
	.db 0x0d, 0x0a
	.db "Select a pattern (1-8): ", 0x00, 0x00

upload:
	.db 0x0d, 0x0a
	.db "Enter pattern using '0' for 'off' and   ", 0x0d, 0x0a
	.db "'1' for 'on':", 0x0d, 0x0a, 0x00

upload_save:
	.db 0x0d, 0x0a
	.db "Saving pattern ...", 0x0d, 0x0a

upload_done:
	.db " done!", 0x0d, 0x0a, 0x00, 0x00

;===============================================================================
